import sys
from pathlib import Path
from typing import Optional

import click
from pydantic import TypeAdapter

from dagster_components import ComponentRegistry
from dagster_components.core.deployment import (
    CodeLocationProjectContext,
    is_inside_code_location_project,
)
from dagster_components.generate import generate_component_instance
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


@click.group(name="generate")
def generate_cli() -> None:
    """Commands for generating Dagster components and related entities."""


@generate_cli.command(name="component")
@click.argument("component_type", type=str)
@click.argument("component_name", type=str)
@click.option("--json-params", type=str, default=None)
@click.pass_context
def generate_component_command(
    ctx: click.Context,
    component_type: str,
    component_name: str,
    json_params: Optional[str],
) -> None:
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    if not is_inside_code_location_project(Path.cwd()):
        click.echo(
            click.style(
                "This command must be run inside a Dagster code location project.", fg="red"
            )
        )
        sys.exit(1)

    context = CodeLocationProjectContext.from_path(
        Path.cwd(),
        ComponentRegistry.from_entry_point_discovery(builtin_component_lib=builtin_component_lib),
    )
    if not context.has_component_type(component_type):
        click.echo(
            click.style(f"No component type `{component_type}` could be resolved.", fg="red")
        )
        sys.exit(1)

    component_type_cls = context.get_component_type(component_type)
    if json_params:
        generate_params_schema = component_type_cls.generate_params_schema
        generate_params = TypeAdapter(generate_params_schema).validate_json(json_params)
    else:
        generate_params = {}

    generate_component_instance(
        context.component_instances_root_path,
        component_name,
        component_type_cls,
        component_type,
        generate_params,
    )



from dagster_components.core.component import (
    Component as Component,
    ComponentGenerateRequest as ComponentGenerateRequest,
    ComponentLoadContext as ComponentLoadContext,
    ComponentRegistry as ComponentRegistry,
    component as component,
)
from dagster_components.core.component_defs_builder import (
    build_defs_from_toplevel_components_folder as build_defs_from_toplevel_components_folder,
)
from dagster_components.version import __version__ as __version__



import os
from pathlib import Path
from typing import Final, Iterable, Tuple, Type

import tomli
from dagster._core.errors import DagsterError
from typing_extensions import Self

from dagster_components.core.component import Component, ComponentRegistry

# Code location
_CODE_LOCATION_CUSTOM_COMPONENTS_DIR: Final = "lib"
_CODE_LOCATION_COMPONENT_INSTANCES_DIR: Final = "components"


def is_inside_code_location_project(path: Path) -> bool:
    try:
        _resolve_code_location_root_path(path)
        return True
    except DagsterError:
        return False


def _resolve_code_location_root_path(path: Path) -> Path:
    current_path = path.absolute()
    while not _is_code_location_root(current_path):
        current_path = current_path.parent
        if str(current_path) == "/":
            raise DagsterError("Cannot find code location root")
    return current_path


def _is_code_location_root(path: Path) -> bool:
    if (path / "pyproject.toml").exists():
        with open(path / "pyproject.toml") as f:
            toml = tomli.loads(f.read())
            return bool(toml.get("tool", {}).get("dagster"))
    return False


class CodeLocationProjectContext:
    @classmethod
    def from_path(cls, path: Path, component_registry: "ComponentRegistry") -> Self:
        root_path = _resolve_code_location_root_path(path)
        return cls(
            root_path=str(root_path),
            name=os.path.basename(root_path),
            component_registry=component_registry,
        )

    def __init__(
        self,
        root_path: str,
        name: str,
        component_registry: "ComponentRegistry",
    ):
        self._root_path = root_path
        self._name = name
        self._component_registry = component_registry

    @property
    def component_types_root_path(self) -> str:
        return os.path.join(self._root_path, self._name, _CODE_LOCATION_CUSTOM_COMPONENTS_DIR)

    @property
    def component_types_root_module(self) -> str:
        return f"{self._name}.{_CODE_LOCATION_CUSTOM_COMPONENTS_DIR}"

    @property
    def component_registry(self) -> "ComponentRegistry":
        return self._component_registry

    def has_component_type(self, name: str) -> bool:
        return self._component_registry.has(name)

    def get_component_type(self, name: str) -> Type[Component]:
        if not self.has_component_type(name):
            raise DagsterError(f"No component type named {name}")
        return self._component_registry.get(name)

    def list_component_types(self) -> Iterable[Tuple[str, Type[Component]]]:
        for key in sorted(self._component_registry.keys()):
            yield key, self._component_registry.get(key)

    def get_component_instance_path(self, name: str) -> str:
        if name not in self.component_instances:
            raise DagsterError(f"No component instance named {name}")
        return os.path.join(self.component_instances_root_path, name)

    @property
    def component_instances_root_path(self) -> str:
        return os.path.join(self._root_path, self._name, _CODE_LOCATION_COMPONENT_INSTANCES_DIR)

    @property
    def component_instances(self) -> Iterable[str]:
        return sorted(
            os.listdir(
                os.path.join(self._root_path, self._name, _CODE_LOCATION_COMPONENT_INSTANCES_DIR)
            )
        )

    def has_component_instance(self, name: str) -> bool:
        return os.path.exists(
            os.path.join(self._root_path, self._name, _CODE_LOCATION_COMPONENT_INSTANCES_DIR, name)
        )



import copy
import dataclasses
import importlib
import importlib.metadata
import inspect
import sys
import textwrap
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import (
    Any,
    ClassVar,
    Dict,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypedDict,
    TypeVar,
)

import click
from dagster import _check as check
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.errors import DagsterError
from dagster._record import record
from dagster._utils import pushd, snakecase
from pydantic import TypeAdapter
from typing_extensions import Self

from dagster_components.core.component_rendering import TemplatedValueResolver, preprocess_value
from dagster_components.utils import ensure_dagster_components_tests_import


class ComponentDeclNode: ...


@record
class ComponentGenerateRequest:
    component_type_name: str
    component_instance_root_path: Path


class Component(ABC):
    name: ClassVar[Optional[str]] = None
    params_schema: ClassVar = None
    generate_params_schema: ClassVar = None

    @classmethod
    def generate_files(cls, request: ComponentGenerateRequest, params: Any) -> None: ...

    @abstractmethod
    def build_defs(self, context: "ComponentLoadContext") -> Definitions: ...

    @classmethod
    @abstractmethod
    def load(cls, context: "ComponentLoadContext") -> Self: ...

    @classmethod
    def get_metadata(cls) -> "ComponentInternalMetadata":
        docstring = cls.__doc__
        clean_docstring = _clean_docstring(docstring) if docstring else None

        return {
            "summary": clean_docstring.split("\n\n")[0] if clean_docstring else None,
            "description": clean_docstring if clean_docstring else None,
            "generate_params_schema": cls.generate_params_schema.schema()
            if cls.generate_params_schema
            else None,
            "component_params_schema": cls.params_schema.schema() if cls.params_schema else None,
        }

    @classmethod
    def get_description(cls) -> Optional[str]:
        return inspect.getdoc(cls)


def _clean_docstring(docstring: str) -> str:
    lines = docstring.strip().splitlines()
    first_line = lines[0]
    if len(lines) == 1:
        return first_line
    else:
        rest = textwrap.dedent("\n".join(lines[1:]))
        return f"{first_line}\n{rest}"


def _get_click_cli_help(command: click.Command) -> str:
    with click.Context(command) as ctx:
        formatter = click.formatting.HelpFormatter()
        param_records = [
            p.get_help_record(ctx) for p in command.get_params(ctx) if p.name != "help"
        ]
        formatter.write_dl([pr for pr in param_records if pr])
        return formatter.getvalue()


class ComponentInternalMetadata(TypedDict):
    summary: Optional[str]
    description: Optional[str]
    generate_params_schema: Optional[Any]  # json schema
    component_params_schema: Optional[Any]  # json schema


class ComponentMetadata(ComponentInternalMetadata):
    name: str
    package: str


def get_entry_points_from_python_environment(group: str) -> Sequence[importlib.metadata.EntryPoint]:
    if sys.version_info >= (3, 10):
        return importlib.metadata.entry_points(group=group)
    else:
        return importlib.metadata.entry_points().get(group, [])


COMPONENTS_ENTRY_POINT_GROUP = "dagster.components"
BUILTIN_COMPONENTS_ENTRY_POINT_BASE = "dagster_components"
BUILTIN_PUBLISHED_COMPONENT_ENTRY_POINT = BUILTIN_COMPONENTS_ENTRY_POINT_BASE
BUILTIN_TEST_COMPONENT_ENTRY_POINT = ".".join([BUILTIN_COMPONENTS_ENTRY_POINT_BASE, "test"])


class ComponentRegistry:
    @classmethod
    def from_entry_point_discovery(
        cls, builtin_component_lib: str = BUILTIN_PUBLISHED_COMPONENT_ENTRY_POINT
    ) -> "ComponentRegistry":
        """Discover components registered in the Python environment via the `dagster_components` entry point group.

        `dagster-components` itself registers multiple component entry points. We call these
        "builtin" component libraries. The `dagster_components` entry point resolves to published
        components and is loaded by default. Other entry points resolve to various sets of test
        components. This method will only ever load one builtin component library.

        Args:
            builtin-component-lib (str): Specifies the builtin components library to load. Builtin
            copmonents libraries are defined under entry points with names matching the pattern
            `dagster_components*`. Only one builtin component library can be loaded at a time.
            Defaults to `dagster_components`, the standard set of published components.
        """
        components: Dict[str, Type[Component]] = {}
        for entry_point in get_entry_points_from_python_environment(COMPONENTS_ENTRY_POINT_GROUP):
            # Skip builtin entry points that are not the specified builtin component library.
            if (
                entry_point.name.startswith(BUILTIN_COMPONENTS_ENTRY_POINT_BASE)
                and not entry_point.name == builtin_component_lib
            ):
                continue
            elif entry_point.name == BUILTIN_TEST_COMPONENT_ENTRY_POINT:
                if builtin_component_lib:
                    ensure_dagster_components_tests_import()
                else:
                    continue

            root_module = entry_point.load()
            if not isinstance(root_module, ModuleType):
                raise DagsterError(
                    f"Invalid entry point {entry_point.name} in group {COMPONENTS_ENTRY_POINT_GROUP}. "
                    f"Value expected to be a module, got {root_module}."
                )
            for component in get_registered_components_in_module(root_module):
                key = f"{entry_point.name}.{get_component_name(component)}"
                components[key] = component

        return cls(components)

    def __init__(self, components: Dict[str, Type[Component]]):
        self._components: Dict[str, Type[Component]] = copy.copy(components)

    @staticmethod
    def empty() -> "ComponentRegistry":
        return ComponentRegistry({})

    def register(self, name: str, component: Type[Component]) -> None:
        if name in self._components:
            raise DagsterError(f"There is an existing component registered under {name}")
        self._components[name] = component

    def has(self, name: str) -> bool:
        return name in self._components

    def get(self, name: str) -> Type[Component]:
        return self._components[name]

    def keys(self) -> Iterable[str]:
        return self._components.keys()

    def __repr__(self) -> str:
        return f"<ComponentRegistry {list(self._components.keys())}>"


def get_registered_components_in_module(module: ModuleType) -> Iterable[Type[Component]]:
    from dagster._core.definitions.load_assets_from_modules import find_subclasses_in_module

    for component in find_subclasses_in_module(module, (Component,)):
        if is_registered_component(component):
            yield component


T = TypeVar("T")


@dataclass
class ComponentLoadContext:
    resources: Mapping[str, object]
    registry: ComponentRegistry
    decl_node: Optional[ComponentDeclNode]
    templated_value_resolver: TemplatedValueResolver

    @staticmethod
    def for_test(
        *,
        resources: Optional[Mapping[str, object]] = None,
        registry: Optional[ComponentRegistry] = None,
        decl_node: Optional[ComponentDeclNode] = None,
    ) -> "ComponentLoadContext":
        return ComponentLoadContext(
            resources=resources or {},
            registry=registry or ComponentRegistry.empty(),
            decl_node=decl_node,
            templated_value_resolver=TemplatedValueResolver.default(),
        )

    @property
    def path(self) -> Path:
        from dagster_components.core.component_decl_builder import YamlComponentDecl

        if not isinstance(self.decl_node, YamlComponentDecl):
            check.failed(f"Unsupported decl_node type {type(self.decl_node)}")

        return self.decl_node.path

    def for_decl_node(self, decl_node: ComponentDeclNode) -> "ComponentLoadContext":
        return dataclasses.replace(self, decl_node=decl_node)

    def _raw_params(self) -> Optional[Mapping[str, Any]]:
        from dagster_components.core.component_decl_builder import YamlComponentDecl

        if not isinstance(self.decl_node, YamlComponentDecl):
            check.failed(f"Unsupported decl_node type {type(self.decl_node)}")
        return self.decl_node.component_file_model.params

    def load_params(self, params_schema: Type[T]) -> T:
        with pushd(str(self.path)):
            preprocessed_params = preprocess_value(
                self.templated_value_resolver, self._raw_params(), params_schema
            )
            return TypeAdapter(params_schema).validate_python(preprocessed_params)


COMPONENT_REGISTRY_KEY_ATTR = "__dagster_component_registry_key"


def component(
    cls: Optional[Type[Component]] = None,
    *,
    name: Optional[str] = None,
) -> Any:
    if cls is None:

        def wrapper(actual_cls: Type[Component]) -> Type[Component]:
            check.inst_param(actual_cls, "actual_cls", type)
            setattr(
                actual_cls,
                COMPONENT_REGISTRY_KEY_ATTR,
                name or snakecase(actual_cls.__name__),
            )
            return actual_cls

        return wrapper
    else:
        # called without params
        check.inst_param(cls, "cls", type)
        setattr(cls, COMPONENT_REGISTRY_KEY_ATTR, name or snakecase(cls.__name__))
        return cls


def is_registered_component(cls: Type) -> bool:
    return hasattr(cls, COMPONENT_REGISTRY_KEY_ATTR)


def get_component_name(component_type: Type[Component]) -> str:
    check.param_invariant(
        is_registered_component(component_type),
        "component_type",
        "Expected a registered component. Use @component to register a component.",
    )
    return getattr(component_type, COMPONENT_REGISTRY_KEY_ATTR)



from abc import ABC, abstractmethod
from typing import Annotated, Any, Dict, Literal, Mapping, Optional, Sequence, Union

from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.asset_spec import AssetSpec, map_asset_specs
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster._record import replace
from pydantic import BaseModel, Field


class OpSpecBaseModel(BaseModel):
    name: Optional[str] = None
    tags: Optional[Dict[str, str]] = None


class AutomationConditionModel(BaseModel):
    type: str
    params: Mapping[str, Any] = {}

    def to_automation_condition(self) -> AutomationCondition:
        return getattr(AutomationCondition, self.type)(**self.params)


class AssetSpecProcessor(ABC, BaseModel):
    target: str = "*"
    description: Optional[str] = None
    metadata: Optional[Mapping[str, Any]] = None
    group_name: Optional[str] = None
    tags: Optional[Mapping[str, str]] = None
    automation_condition: Optional[AutomationConditionModel] = None

    def _attributes(self) -> Mapping[str, Any]:
        return {
            **self.model_dump(exclude={"target", "operation"}, exclude_unset=True),
            **{
                "automation_condition": self.automation_condition.to_automation_condition()
                if self.automation_condition
                else None
            },
        }

    @abstractmethod
    def _apply_to_spec(self, spec: AssetSpec) -> AssetSpec: ...

    def apply(self, defs: Definitions) -> Definitions:
        target_selection = AssetSelection.from_string(self.target, include_sources=True)
        target_keys = target_selection.resolve(defs.get_asset_graph())

        mappable = [d for d in defs.assets or [] if isinstance(d, (AssetsDefinition, AssetSpec))]
        mapped_assets = map_asset_specs(
            lambda spec: self._apply_to_spec(spec) if spec.key in target_keys else spec, mappable
        )

        assets = [
            *mapped_assets,
            *[d for d in defs.assets or [] if not isinstance(d, (AssetsDefinition, AssetSpec))],
        ]
        return replace(defs, assets=assets)


class MergeAttributes(AssetSpecProcessor):
    # default operation is "merge"
    operation: Literal["merge"] = "merge"

    def _apply_to_spec(self, spec: AssetSpec) -> AssetSpec:
        attributes = self._attributes()
        mergeable_attributes = {"metadata", "tags"}
        merge_attributes = {k: v for k, v in attributes.items() if k in mergeable_attributes}
        replace_attributes = {k: v for k, v in attributes.items() if k not in mergeable_attributes}
        return spec.merge_attributes(**merge_attributes).replace_attributes(**replace_attributes)


class ReplaceAttributes(AssetSpecProcessor):
    # operation must be set explicitly
    operation: Literal["replace"]

    def _apply_to_spec(self, spec: AssetSpec) -> AssetSpec:
        return spec.replace_attributes(**self._attributes())


AssetAttributes = Sequence[
    Annotated[
        Union[MergeAttributes, ReplaceAttributes],
        Field(union_mode="left_to_right"),
    ]
]



from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, Union

from dagster._record import record
from dagster._utils.pydantic_yaml import parse_yaml_file_to_pydantic
from pydantic import BaseModel

from dagster_components.core.component import ComponentDeclNode


class ComponentFileModel(BaseModel):
    type: str
    params: Optional[Mapping[str, Any]] = None


@record
class YamlComponentDecl(ComponentDeclNode):
    path: Path
    component_file_model: ComponentFileModel


@record
class ComponentFolder(ComponentDeclNode):
    path: Path
    sub_decls: Sequence[Union[YamlComponentDecl, "ComponentFolder"]]


def path_to_decl_node(path: Path) -> Optional[ComponentDeclNode]:
    # right now, we only support two types of components, both of which are folders
    # if the folder contains a component.yaml file, it's a component instance
    # otherwise, it's a folder containing sub-components

    if not path.is_dir():
        return None

    component_path = path / "component.yaml"

    if component_path.exists():
        component_file_model = parse_yaml_file_to_pydantic(
            ComponentFileModel, component_path.read_text(), str(path)
        )
        return YamlComponentDecl(path=path, component_file_model=component_file_model)

    subs = []
    for subpath in path.iterdir():
        component = path_to_decl_node(subpath)
        if component:
            subs.append(component)

    return ComponentFolder(path=path, sub_decls=subs) if subs else None



import os
from pathlib import Path
from typing import Any, Mapping, Optional, Type

import click
import yaml
from dagster._utils import mkdir_p

from dagster_components.core.component import Component, ComponentGenerateRequest


class ComponentDumper(yaml.Dumper):
    def write_line_break(self) -> None:
        # add an extra line break between top-level keys
        if self.indent == 0:
            super().write_line_break()
        super().write_line_break()


def generate_component_yaml(
    request: ComponentGenerateRequest, component_params: Optional[Mapping[str, Any]]
) -> None:
    with open(request.component_instance_root_path / "component.yaml", "w") as f:
        component_data = {"type": request.component_type_name, "params": component_params or {}}
        yaml.dump(
            component_data, f, Dumper=ComponentDumper, sort_keys=False, default_flow_style=False
        )


def generate_component_instance(
    root_path: str,
    name: str,
    component_type: Type[Component],
    component_type_name: str,
    generate_params: Mapping[str, Any],
) -> None:
    component_instance_root_path = Path(os.path.join(root_path, name))
    click.echo(f"Creating a Dagster component instance folder at {component_instance_root_path}.")
    mkdir_p(str(component_instance_root_path))
    component_type.generate_files(
        ComponentGenerateRequest(
            component_type_name=component_type_name,
            component_instance_root_path=component_instance_root_path,
        ),
        generate_params,
    )

    component_yaml_path = component_instance_root_path / "component.yaml"

    if not component_yaml_path.exists():
        raise Exception(
            f"Currently all components require a component.yaml file. Please ensure your implementation of generate_files writes this file at {component_yaml_path}."
        )



import os
from pathlib import Path
from typing import Any, Iterator, Mapping, Optional, Sequence

import dagster._check as check
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._utils import pushd
from dagster_dbt import DagsterDbtTranslator, DbtCliResource, DbtProject, dbt_assets
from dbt.cli.main import dbtRunner
from pydantic import BaseModel, Field
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import (
    ComponentGenerateRequest,
    TemplatedValueResolver,
    component,
)
from dagster_components.core.component_rendering import RenderingScope
from dagster_components.core.dsl_schema import AssetAttributes, AssetSpecProcessor, OpSpecBaseModel
from dagster_components.generate import generate_component_yaml


class DbtNodeTranslatorParams(BaseModel):
    key: Optional[str] = None
    group: Optional[str] = None


class DbtProjectParams(BaseModel):
    dbt: DbtCliResource
    op: Optional[OpSpecBaseModel] = None
    translator: Optional[DbtNodeTranslatorParams] = RenderingScope(
        Field(default=None), required_scope={"node"}
    )
    asset_attributes: Optional[AssetAttributes] = None


class DbtGenerateParams(BaseModel):
    init: bool = Field(default=False)
    project_path: Optional[str] = None


class DbtProjectComponentTranslator(DagsterDbtTranslator):
    def __init__(
        self,
        *,
        value_resolver: TemplatedValueResolver,
        translator_params: Optional[DbtNodeTranslatorParams] = None,
    ):
        self.value_resolver = value_resolver
        self.translator_params = translator_params

    def get_asset_key(self, dbt_resource_props: Mapping[str, Any]) -> AssetKey:
        if not self.translator_params or not self.translator_params.key:
            return super().get_asset_key(dbt_resource_props)

        return AssetKey.from_user_string(
            self.value_resolver.with_context(node=dbt_resource_props).resolve(
                self.translator_params.key
            )
        )

    def get_group_name(self, dbt_resource_props) -> Optional[str]:
        if not self.translator_params or not self.translator_params.group:
            return super().get_group_name(dbt_resource_props)

        return self.value_resolver.with_context(node=dbt_resource_props).resolve(
            self.translator_params.group
        )


@component(name="dbt_project")
class DbtProjectComponent(Component):
    params_schema = DbtProjectParams
    generate_params_schema = DbtGenerateParams

    def __init__(
        self,
        dbt_resource: DbtCliResource,
        op_spec: Optional[OpSpecBaseModel],
        dbt_translator: Optional[DagsterDbtTranslator],
        asset_processors: Sequence[AssetSpecProcessor],
    ):
        self.dbt_resource = dbt_resource
        self.op_spec = op_spec
        self.dbt_translator = dbt_translator
        self.asset_processors = asset_processors

    @classmethod
    def load(cls, context: ComponentLoadContext) -> Self:
        # all paths should be resolved relative to the directory we're in
        with pushd(str(context.path)):
            loaded_params = context.load_params(cls.params_schema)

        return cls(
            dbt_resource=loaded_params.dbt,
            op_spec=loaded_params.op,
            dbt_translator=DbtProjectComponentTranslator(
                translator_params=loaded_params.translator,
                value_resolver=context.templated_value_resolver,
            ),
            asset_processors=loaded_params.asset_attributes or [],
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        project = DbtProject(self.dbt_resource.project_dir)
        project.prepare_if_dev()

        @dbt_assets(
            manifest=project.manifest_path,
            project=project,
            name=self.op_spec.name if self.op_spec else project.name,
            op_tags=self.op_spec.tags if self.op_spec else None,
            dagster_dbt_translator=self.dbt_translator,
        )
        def _fn(context: AssetExecutionContext):
            yield from self.execute(context=context, dbt=self.dbt_resource)

        defs = Definitions(assets=[_fn])
        for transform in self.asset_processors:
            defs = transform.apply(defs)
        return defs

    @classmethod
    def generate_files(cls, request: ComponentGenerateRequest, params: DbtGenerateParams) -> None:
        cwd = os.getcwd()
        if params.project_path:
            # NOTE: CWD is not set "correctly" above so we prepend "../../.." as a temporary hack to
            # make sure the path is right.
            relative_path = os.path.join(
                "../../../", os.path.relpath(params.project_path, start=cwd)
            )
        elif params.init:
            dbtRunner().invoke(["init"])
            subpaths = [
                path for path in Path(cwd).iterdir() if path.is_dir() and path.name != "logs"
            ]
            check.invariant(len(subpaths) == 1, "Expected exactly one subpath to be created.")
            # this path should be relative to this directory
            relative_path = subpaths[0].name
        else:
            relative_path = None

        generate_component_yaml(request, {"dbt": {"project_dir": relative_path}})

    def execute(self, context: AssetExecutionContext, dbt: DbtCliResource) -> Iterator:
        yield from dbt.cli(["build"], context=context).stream()



import json
import sys
from pathlib import Path
from typing import Any, Dict

import click

from dagster_components.core.component import ComponentMetadata, ComponentRegistry
from dagster_components.core.deployment import (
    CodeLocationProjectContext,
    is_inside_code_location_project,
)
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


@click.group(name="generate")
def list_cli():
    """Commands for listing Dagster components and related entities."""


@list_cli.command(name="component-types")
@click.pass_context
def list_component_types_command(ctx: click.Context) -> None:
    """List registered Dagster components."""
    builtin_component_lib = ctx.obj.get(CLI_BUILTIN_COMPONENT_LIB_KEY, False)
    if not is_inside_code_location_project(Path.cwd()):
        click.echo(
            click.style(
                "This command must be run inside a Dagster code location project.", fg="red"
            )
        )
        sys.exit(1)

    context = CodeLocationProjectContext.from_path(
        Path.cwd(),
        ComponentRegistry.from_entry_point_discovery(builtin_component_lib=builtin_component_lib),
    )
    output: Dict[str, Any] = {}
    for key, component_type in context.list_component_types():
        package, name = key.rsplit(".", 1)
        output[key] = ComponentMetadata(
            name=name,
            package=package,
            **component_type.get_metadata(),
        )
    click.echo(json.dumps(output))



import click
from dagster.version import __version__

from dagster_components.cli.generate import generate_cli
from dagster_components.cli.list import list_cli
from dagster_components.core.component import BUILTIN_PUBLISHED_COMPONENT_ENTRY_POINT
from dagster_components.utils import CLI_BUILTIN_COMPONENT_LIB_KEY


def create_dagster_components_cli():
    commands = {
        "generate": generate_cli,
        "list": list_cli,
    }

    @click.group(
        commands=commands,
        context_settings={"max_content_width": 120, "help_option_names": ["-h", "--help"]},
    )
    @click.option(
        "--builtin-component-lib",
        type=str,
        default=BUILTIN_PUBLISHED_COMPONENT_ENTRY_POINT,
        help="Specify the builitin component library to load.",
    )
    @click.version_option(__version__, "--version", "-v")
    @click.pass_context
    def group(ctx: click.Context, builtin_component_lib: str):
        """CLI tools for working with Dagster."""
        ctx.ensure_object(dict)
        ctx.obj[CLI_BUILTIN_COMPONENT_LIB_KEY] = builtin_component_lib

    return group


ENV_PREFIX = "DG_CLI"
cli = create_dagster_components_cli()


def main():
    cli(auto_envvar_prefix=ENV_PREFIX)



import os
from pathlib import Path
from typing import Any, Iterator, Optional, Sequence, Union

import yaml
from dagster._core.definitions.definitions_class import Definitions
from dagster._core.definitions.events import AssetMaterialization
from dagster._core.definitions.result import MaterializeResult
from dagster_embedded_elt.sling import SlingResource, sling_assets
from dagster_embedded_elt.sling.resources import AssetExecutionContext
from pydantic import BaseModel
from typing_extensions import Self

from dagster_components import Component, ComponentLoadContext
from dagster_components.core.component import ComponentGenerateRequest, component
from dagster_components.core.dsl_schema import AssetAttributes, AssetSpecProcessor, OpSpecBaseModel
from dagster_components.generate import generate_component_yaml


class SlingReplicationParams(BaseModel):
    sling: Optional[SlingResource] = None
    op: Optional[OpSpecBaseModel] = None
    asset_attributes: Optional[AssetAttributes] = None


@component(name="sling_replication")
class SlingReplicationComponent(Component):
    params_schema = SlingReplicationParams

    def __init__(
        self,
        dirpath: Path,
        resource: SlingResource,
        op_spec: Optional[OpSpecBaseModel],
        asset_processors: Sequence[AssetSpecProcessor],
    ):
        self.dirpath = dirpath
        self.resource = resource
        self.op_spec = op_spec
        self.asset_processors = asset_processors

    @classmethod
    def load(cls, context: ComponentLoadContext) -> Self:
        loaded_params = context.load_params(cls.params_schema)
        return cls(
            dirpath=context.path,
            resource=loaded_params.sling or SlingResource(),
            op_spec=loaded_params.op,
            asset_processors=loaded_params.asset_attributes or [],
        )

    def build_defs(self, context: ComponentLoadContext) -> Definitions:
        @sling_assets(
            name=self.op_spec.name if self.op_spec else self.dirpath.stem,
            op_tags=self.op_spec.tags if self.op_spec else {},
            replication_config=self.dirpath / "replication.yaml",
        )
        def _fn(context: AssetExecutionContext, sling: SlingResource):
            yield from self.execute(context=context, sling=sling)

        defs = Definitions(assets=[_fn], resources={"sling": self.resource})
        for transform in self.asset_processors:
            defs = transform.apply(defs)
        return defs

    @classmethod
    def generate_files(cls, request: ComponentGenerateRequest, params: Any) -> None:
        generate_component_yaml(request, params)
        replication_path = Path(os.getcwd()) / "replication.yaml"
        with open(replication_path, "w") as f:
            yaml.dump(
                {"source": {}, "target": {}, "streams": {}},
                f,
            )

    def execute(
        self, context: AssetExecutionContext, sling: SlingResource
    ) -> Iterator[Union[AssetMaterialization, MaterializeResult]]:
        yield from sling.replicate(context=context)



import sys
from pathlib import Path

CLI_BUILTIN_COMPONENT_LIB_KEY = "builtin_component_lib"


def ensure_dagster_components_tests_import() -> None:
    from dagster_components import __file__ as dagster_components_init_py

    dagster_components_package_root = (Path(dagster_components_init_py) / ".." / "..").resolve()
    assert (
        dagster_components_package_root / "dagster_components_tests"
    ).exists(), "Could not find dagster_components_tests where expected"
    sys.path.append(dagster_components_package_root.as_posix())



import shutil
from pathlib import Path
from typing import TYPE_CHECKING, Any, Mapping, Optional, Sequence

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_spec import AssetSpec
from dagster._core.definitions.assets import AssetsDefinition
from dagster._core.definitions.decorators.asset_decorator import multi_asset
from dagster._core.execution.context.asset_execution_context import AssetExecutionContext
from dagster._core.pipes.subprocess import PipesSubprocessClient
from dagster._utils.warnings import suppress_dagster_warnings
from pydantic import BaseModel

from dagster_components.core.component import Component, ComponentLoadContext, component
from dagster_components.core.dsl_schema import AutomationConditionModel

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


class AssetSpecModel(BaseModel):
    key: str
    deps: Sequence[str] = []
    description: Optional[str] = None
    metadata: Mapping[str, Any] = {}
    group_name: Optional[str] = None
    skippable: bool = False
    code_version: Optional[str] = None
    owners: Sequence[str] = []
    tags: Mapping[str, str] = {}
    automation_condition: Optional[AutomationConditionModel] = None

    @suppress_dagster_warnings
    def to_asset_spec(self) -> AssetSpec:
        return AssetSpec(
            **{
                **self.__dict__,
                "key": AssetKey.from_user_string(self.key),
                "automation_condition": self.automation_condition.to_automation_condition()
                if self.automation_condition
                else None,
            },
        )


class PipesSubprocessScriptParams(BaseModel):
    path: str
    assets: Sequence[AssetSpecModel]


class PipesSubprocessScriptCollectionParams(BaseModel):
    scripts: Sequence[PipesSubprocessScriptParams]


@component(name="pipes_subprocess_script_collection")
class PipesSubprocessScriptCollection(Component):
    """Assets that wrap Python scripts executed with Dagster's PipesSubprocessClient."""

    params_schema = PipesSubprocessScriptCollectionParams

    def __init__(self, dirpath: Path, path_specs: Mapping[Path, Sequence[AssetSpec]]):
        self.dirpath = dirpath
        # mapping from the script name (e.g. /path/to/script_abc.py -> script_abc)
        # to the specs it produces
        self.path_specs = path_specs

    @staticmethod
    def introspect_from_path(path: Path) -> "PipesSubprocessScriptCollection":
        path_specs = {path: [AssetSpec(path.stem)] for path in list(path.rglob("*.py"))}
        return PipesSubprocessScriptCollection(dirpath=path, path_specs=path_specs)

    @classmethod
    def load(cls, context: ComponentLoadContext) -> "PipesSubprocessScriptCollection":
        loaded_params = context.load_params(cls.params_schema)

        path_specs = {}
        for script in loaded_params.scripts:
            script_path = context.path / script.path
            if not script_path.exists():
                raise FileNotFoundError(f"Script {script_path} does not exist")
            path_specs[script_path] = [spec.to_asset_spec() for spec in script.assets]

        return cls(dirpath=context.path, path_specs=path_specs)

    def build_defs(self, load_context: "ComponentLoadContext") -> "Definitions":
        from dagster._core.definitions.definitions_class import Definitions

        return Definitions(
            assets=[self._create_asset_def(path, specs) for path, specs in self.path_specs.items()],
            resources={"pipes_client": PipesSubprocessClient()},
        )

    def _create_asset_def(self, path: Path, specs: Sequence[AssetSpec]) -> AssetsDefinition:
        # TODO: allow name paraeterization
        @multi_asset(specs=specs, name=f"script_{path.stem}")
        def _asset(context: AssetExecutionContext, pipes_client: PipesSubprocessClient):
            cmd = [shutil.which("python"), path]
            return pipes_client.run(command=cmd, context=context).get_results()

        return _asset



import importlib.util

_has_dagster_dbt = importlib.util.find_spec("dagster_dbt") is not None
_has_dagster_embedded_elt = importlib.util.find_spec("dagster_embedded_elt") is not None

if _has_dagster_dbt:
    from dagster_components.lib.dbt_project import DbtProjectComponent as DbtProjectComponent

if _has_dagster_embedded_elt:
    from dagster_components.lib.sling_replication import (
        SlingReplicationComponent as SlingReplicationComponent,
    )

from dagster_components.lib.pipes_subprocess_script_collection import (
    PipesSubprocessScriptCollection as PipesSubprocessScriptCollection,
)



import json
import os
from typing import AbstractSet, Any, Mapping, Optional, Sequence, Type, TypeVar, Union

import dagster._check as check
from dagster._record import record
from jinja2 import Template
from pydantic import BaseModel, Field
from pydantic.fields import FieldInfo

T = TypeVar("T")

REF_BASE = "#/$defs/"
REF_TEMPLATE = f"{REF_BASE}{{model}}"

CONTEXT_KEY = "required_rendering_scope"


def RenderingScope(field: Optional[FieldInfo] = None, *, required_scope: AbstractSet[str]) -> Any:
    """Defines a Pydantic Field that requires a specific scope to be available before rendering.

    Examples:
    ```python
    class Schema(BaseModel):
        a: str = RenderingScope(required_scope={"foo", "bar"})
        b: Optional[int] = RenderingScope(Field(default=None), required_scope={"baz"})
    ```
    """
    return FieldInfo.merge_field_infos(
        field or Field(), Field(json_schema_extra={CONTEXT_KEY: json.dumps(list(required_scope))})
    )


def get_required_rendering_context(subschema: Mapping[str, Any]) -> Optional[AbstractSet[str]]:
    raw = check.opt_inst(subschema.get(CONTEXT_KEY), str)
    return set(json.loads(raw)) if raw else None


def _env(key: str) -> Optional[str]:
    return os.environ.get(key)


@record
class TemplatedValueResolver:
    context: Mapping[str, Any]

    @staticmethod
    def default() -> "TemplatedValueResolver":
        return TemplatedValueResolver(context={"env": _env})

    def with_context(self, **additional_context) -> "TemplatedValueResolver":
        return TemplatedValueResolver(context={**self.context, **additional_context})

    def resolve(self, val: str) -> str:
        return Template(val).render(**self.context)


def _should_render(
    valpath: Sequence[Union[str, int]], json_schema: Mapping[str, Any], subschema: Mapping[str, Any]
) -> bool:
    # List[ComplexType] (e.g.) will contain a reference to the complex type schema in the
    # top-level $defs, so we dereference it here.
    if "$ref" in subschema:
        subschema = json_schema["$defs"].get(subschema["$ref"][len(REF_BASE) :])

    if get_required_rendering_context(subschema) is not None:
        return False
    elif len(valpath) == 0:
        return True

    # Optional[ComplexType] (e.g.) will contain multiple schemas in the "anyOf" field
    if "anyOf" in subschema:
        return any(_should_render(valpath, json_schema, inner) for inner in subschema["anyOf"])

    el = valpath[0]
    if isinstance(el, str):
        # valpath: ['field']
        # field: X
        inner = subschema.get("properties", {}).get(el)
    elif isinstance(el, int):
        # valpath: ['field', 0]
        # field: List[X]
        inner = subschema.get("items")
    else:
        check.failed(f"Unexpected valpath element: {el}")

    # the path wasn't valid
    if not inner:
        return False

    _, *rest = valpath
    return _should_render(rest, json_schema, inner)


def _render_values(
    value_resolver: TemplatedValueResolver,
    val: Any,
    valpath: Sequence[Union[str, int]],
    json_schema: Optional[Mapping[str, Any]],
) -> Any:
    if json_schema and not _should_render(valpath, json_schema, json_schema):
        return val
    elif isinstance(val, dict):
        return {
            k: _render_values(value_resolver, v, [*valpath, k], json_schema) for k, v in val.items()
        }
    elif isinstance(val, list):
        return [
            _render_values(value_resolver, v, [*valpath, i], json_schema) for i, v in enumerate(val)
        ]
    else:
        return value_resolver.resolve(val)


def preprocess_value(renderer: TemplatedValueResolver, val: T, target_type: Type) -> T:
    """Given a raw value, preprocesses it by rendering any templated values that are not marked as deferred in the target_type's json schema."""
    json_schema = target_type.model_json_schema() if issubclass(target_type, BaseModel) else None
    return _render_values(renderer, val, [], json_schema)



import importlib
import importlib.util
import inspect
from pathlib import Path
from types import ModuleType
from typing import TYPE_CHECKING, List, Mapping, Optional, Sequence, Type

from dagster._utils.warnings import suppress_dagster_warnings

from dagster_components.core.component import (
    Component,
    ComponentLoadContext,
    ComponentRegistry,
    TemplatedValueResolver,
    get_component_name,
    is_registered_component,
)
from dagster_components.core.component_decl_builder import (
    ComponentFolder,
    YamlComponentDecl,
    path_to_decl_node,
)
from dagster_components.core.deployment import CodeLocationProjectContext

if TYPE_CHECKING:
    from dagster._core.definitions.definitions_class import Definitions


def load_module_from_path(module_name, path) -> ModuleType:
    # Create a spec from the file path
    spec = importlib.util.spec_from_file_location(module_name, path)
    if spec is None:
        raise ImportError(f"Cannot create a module spec from path: {path}")

    # Create and load the module
    module = importlib.util.module_from_spec(spec)
    assert spec.loader, "Must have a loader"
    spec.loader.exec_module(module)
    return module


def load_components_from_context(context: ComponentLoadContext) -> Sequence[Component]:
    if isinstance(context.decl_node, YamlComponentDecl):
        component_type = component_type_from_yaml_decl(context.registry, context.decl_node)
        return [component_type.load(context)]
    elif isinstance(context.decl_node, ComponentFolder):
        components = []
        for sub_decl in context.decl_node.sub_decls:
            components.extend(load_components_from_context(context.for_decl_node(sub_decl)))
        return components

    raise NotImplementedError(f"Unknown component type {context.decl_node}")


def component_type_from_yaml_decl(
    registry: ComponentRegistry, decl_node: YamlComponentDecl
) -> Type[Component]:
    parsed_defs = decl_node.component_file_model
    if parsed_defs.type.startswith("."):
        component_registry_key = parsed_defs.type[1:]

        # Iterate over Python files in the folder
        for py_file in decl_node.path.glob("*.py"):
            module_name = py_file.stem

            module = load_module_from_path(module_name, decl_node.path / f"{module_name}.py")

            for _name, obj in inspect.getmembers(module, inspect.isclass):
                assert isinstance(obj, Type)
                if (
                    is_registered_component(obj)
                    and get_component_name(obj) == component_registry_key
                ):
                    return obj

        raise Exception(
            f"Could not find component type {component_registry_key} in {decl_node.path}"
        )

    return registry.get(parsed_defs.type)


def build_components_from_component_folder(
    context: ComponentLoadContext, path: Path
) -> Sequence[Component]:
    component_folder = path_to_decl_node(path)
    assert isinstance(component_folder, ComponentFolder)
    return load_components_from_context(context.for_decl_node(component_folder))


def build_defs_from_component_path(
    path: Path,
    registry: ComponentRegistry,
    resources: Mapping[str, object],
) -> "Definitions":
    """Build a definitions object from a folder within the components hierarchy."""
    decl_node = path_to_decl_node(path=path)
    if not decl_node:
        raise Exception(f"No component found at path {path}")

    context = ComponentLoadContext(
        resources=resources,
        registry=registry,
        decl_node=decl_node,
        templated_value_resolver=TemplatedValueResolver.default(),
    )
    components = load_components_from_context(context)
    return defs_from_components(resources=resources, context=context, components=components)


@suppress_dagster_warnings
def defs_from_components(
    *,
    context: ComponentLoadContext,
    components: Sequence[Component],
    resources: Mapping[str, object],
) -> "Definitions":
    from dagster._core.definitions.definitions_class import Definitions

    return Definitions.merge(
        *[*[c.build_defs(context) for c in components], Definitions(resources=resources)]
    )


# Public method so optional Nones are fine
@suppress_dagster_warnings
def build_defs_from_toplevel_components_folder(
    path: Path,
    resources: Optional[Mapping[str, object]] = None,
    registry: Optional["ComponentRegistry"] = None,
) -> "Definitions":
    """Build a Definitions object from an entire component hierarchy."""
    from dagster._core.definitions.definitions_class import Definitions

    context = CodeLocationProjectContext.from_path(
        path, registry or ComponentRegistry.from_entry_point_discovery()
    )

    all_defs: List[Definitions] = []
    for component in context.component_instances:
        component_path = Path(context.get_component_instance_path(component))
        defs = build_defs_from_component_path(
            path=component_path,
            registry=context.component_registry,
            resources=resources or {},
        )
        all_defs.append(defs)
    return Definitions.merge(*all_defs)
